// Copyright Derek C. Lazenby, 2019

unit TicTacToeMain;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls,
  ExtCtrls, LCLType;

type
  States = ( Playing, Won, Drawn ) ;
  LineRange = 1..8 ;
  LineSet = set of LineRange ;
  CellRange = 1..9 ;
  CellSet = set of CellRange ;

  { TTTTForm }

  TTTTForm = class(TForm)
    Panel1: TPanel;
    Panel2: TPanel;
    Panel3: TPanel;
    Panel4: TPanel;
    Panel5: TPanel;
    Panel6: TPanel;
    Panel7: TPanel;
    Panel8: TPanel;
    Panel9: TPanel;
    procedure FormCreate(Sender: TObject);
    procedure PanelClick(Sender: TObject);
  private
    PlayerX  : boolean;
    Count    : integer;
    Status   : States;
    XChoices,
    OChoices,
    Empty    : CellSet ;
    const
    Winner   : array [boolean] of char = ( 'O', 'X' ) ;
    Lines    : array [LineRange] of CellSet = ( [1,2,3], [4,5,6], [7,8,9],
                                                [1,4,7], [2,5,8], [3,6,9],
                                                [1,5,9], [3,5,7] );
    Possible : array [CellRange] of LineSet = ( [1,4,7], [1,5],     [1,6,8],
                                                [2,4],   [2,5,7,8], [2,6],
                                                [3,4,8], [3,5],     [3,6,7] );
    procedure ClearGrid;
  public

  end;

var
  TTTForm: TTTTForm;

implementation

{$R *.lfm}

{ TTTTForm }

procedure TTTTForm.FormCreate(Sender: TObject);
begin
  ClearGrid ; // well just init the variables actually!
end;

procedure TTTTForm.ClearGrid;
begin
  Panel1.Caption := '' ;
  Panel2.Caption := '' ;
  Panel3.Caption := '' ;
  Panel4.Caption := '' ;
  Panel5.Caption := '' ;
  Panel6.Caption := '' ;
  Panel7.Caption := '' ;
  Panel8.Caption := '' ;
  Panel9.Caption := '' ;
  XChoices := [] ;
  OChoices := [] ;
  Empty := [Low (CellRange)..High (CellRange)] ;
  Status := Playing ;
  PlayerX := true ;
  Count := 0 ;
end;

procedure TTTTForm.PanelClick(Sender: TObject);

  function TestForWin ( Choices : CellSet ; Choice : CellRange ) : boolean ;

    var Line : LineRange ;

  begin
    for Line in Possible [Choice] do // only check relevant lines
      if Lines [Line] <= Choices then // check whole line !
      begin
        TestForWin := true ;
        exit ;
      end ;
    TestForWin := false ;
  end;

  procedure CheckForWin ( var Choices : CellSet ) ;

    var Choice : CellRange ;

  begin
    Choice := TPanel (Sender).Tag ; // each panel (cell) has Tag set to 1 to 9 respectively
    Choices := Choices + [Choice] ; // update var parameter
    if ( Count > 4 ) and TestForWin ( Choices, Choice ) then
      Status := Won ;
  end;

  procedure CheckForDraw6 ;

    var Choice  : CellRange ;

  begin
    // I think it's just rotations of
    // XO-
    // O-X
    // -XO

    // Do we have an empty diagonal?
    if ( Empty = Lines [7] ) or ( Empty = Lines [8] ) then
    begin
      for Choice in Empty  do
        // Can't have potential O win without potential X win, so only check X
        if TestForWin ( XChoices + [Choice], Choice ) then
          exit ;
	  Status := Drawn ;
	end;
  end;

  procedure CheckForDraw7 ;

    var Choice  : CellRange ;
        Empties : array [boolean] of CellRange ;
        i       : boolean ;

  begin
    i := false ;

    // extract the two empty cell numbers
    for Choice in Empty do
    begin
      Empties [i] := Choice ;
      i := not i ;
	end ;

    // see if filling in empty Cells gives a win
    for i := false to true do
    begin
      Choice := Empties [i] ;
      // try O in selected empty cell
      if TestForWin ( OChoices + [Choice], Choice ) then
        exit ;
      // try X in remaining cell
      Choice := Empties [not i] ;
      if TestForWin ( XChoices + [Choice], Choice ) then
        exit ;
	end ;
    // no potential wins found
    Status := Drawn ;
  end;

  procedure CheckForDraw8 ;

    var Choice  : CellRange ;

  begin
    // calculate the only empty cell number
    // i.e. the loop only iterates once,
    //      but it does get us the cell number from the set
    for Choice in Empty do
      // and see if it would not win for X
      if not TestForWin ( XChoices + [Choice], Choice ) then
        Status := Drawn ;
  end;

  procedure GameOver ( Result : string ) ;
  begin
    if QuestionDlg ( 'Game Over', Result + #13#10#13#10'Play again?', mtConfirmation,
                    [mryes, 'Yeah', mrno, 'Nope'], 0) = mrYes then
      ClearGrid
    else
      Close ;
  end ;

begin
  with TPanel ( Sender ) do
  begin
    Empty := Empty - [Tag] ;
    if Caption = '' then
    begin
      Inc ( Count ) ;
      if PlayerX then
      begin
        Caption := 'X' ;
        CheckForWin ( XChoices ) ;
      end
      else
      begin
        Caption := 'O' ;
        CheckForWin ( OChoices ) ;
      end
    end
  end ;
  if Status <> Won then
    case Count of
    6 : CheckForDraw6 ;
    7 : CheckForDraw7 ;
    8 : CheckForDraw8 ;
    9 : Status := Drawn ;
    else // do nothing
    end;
  case Status of
  Playing : PlayerX := not PlayerX ;
  Won     : GameOver ( Winner [PlayerX] + ' Wins !!!' ) ;
  Drawn   : GameOver ( 'Game drawn.' ) ;
  end
end;

end.

